/*
Created By:     Kerri Morgan	2007
Modified By:	Zoe Bukovac		17th April 2011
				Modified to all functions to work with graphs of variable size.
*/


/* This program prints generators for the automorphism group of an
   n-vertex polygon, where n is a number supplied by the user.
   It needs to be linked with nauty.c, nautil.c and naugraph.c.

   This version uses a fixed limit for MAXN.
*/
#include "iso.h"
#include "nauty.h"   /* which includes <stdio.h> */

int isIso(int **A, int **B, int size)
{
    int i,j;
    graph g[MAXN*MAXM];
    graph g2[MAXN*MAXM];
    graph canong[MAXN*MAXM];
    graph canong2[MAXN*MAXM];
    int lab[MAXN],ptn[MAXN],orbits[MAXN];
    int lab2[MAXN],ptn2[MAXN];
    static DEFAULTOPTIONS_GRAPH(options);
    statsblk(stats);
    setword workspace[50*MAXM];

    int n,m;
    set *gv;

	options.defaultptn=TRUE;
	options.getcanon=TRUE;
	options.writeautoms=FALSE; /* write to stdout as NULL */

	    n=size;
            m = (n + WORDSIZE - 1) / WORDSIZE;
            nauty_check(WORDSIZE,m,n,NAUTYVERSIONID);

	 
	for(i=0; i<size; i++)
	{
		gv=GRAPHROW(g,i,m);
          	EMPTYSET(gv,m);
		for(j=0; j<size; j++)
		{
			if(A[i][j]==1)
			{
	 			ADDELEMENT(gv,j); 
			}
		}
	}


	for(i=0; i<size; i++)
	{
		gv=GRAPHROW(g2,i,m);
          	EMPTYSET(gv,m);
		for(j=0; j<size; j++)
		{
			if(B[i][j]==1)
			{
	 			ADDELEMENT(gv,j); 
			}
		}
	}

	    nauty(g,lab,ptn,NULL,orbits,&options,&stats,workspace,50*MAXM,m,n,canong);
            nauty(g2,lab2,ptn2,NULL,orbits,&options,&stats,workspace,50*MAXM,m,n,canong2);
                                            
	if(memcmp(canong, canong2, m*n*sizeof(graph)))
	{
		//printf("NOT ISOMORPHIC\n");
		return 0;
	}
	else
	{
		//printf("ISOMORPHIC\n");
		return 1;
	}
}
/******************************************************************************
** Function:   isIso_H_add_e
** Description: Function checks if G is isomorphic to H+e where e is some edge
** Inputs:      G - the graph
**              size - size of graph
**		H - the second graph
**		h_size - size of H
**		edge - storage to return the edge e such that G~H+e
** Return Value: 1 if G ~ H+e otherwise 0
** Side Effects: updates edge
** Created:     Kerri Morgan    9th May, 2007
** Modified:	Zoe Bukovac		17th April 2011
				Removed fixed size of graphs
** Comments:
** Acknowledgements:
******************************************************************************/
int isIso_H_add_e(int **G, int size, int **H, int h_size, int edge[2]) 
{
	int i, j;

	edge[0]=-1;
	edge[1]=-1;

	if(size != h_size)
		return 0;


	/* for each edge not in graph */
	for(i=0; i<size; i++)
	{
		for(j=i+1; j<size; j++)
		{
			if(H[i][j]==0)
			{
				H[i][j]=1;
				H[j][i]=1;
				if (isIso(G, H, size))
				{
					edge[0]=i;
					edge[1]=j;
					H[i][j]=0;
					H[j][i]=0;
					return 1;
				}
				else
				{
					H[i][j]=0;
					H[j][i]=0;
				}
			}
		}
	}
	return 0;
}
/******************************************************************************
** Function:   isIso_H_efg_plus
** Description: Function checks if G is isomorphic to H-e+f+g where
**		e is  in H2 and f and g are not
** Inputs:      G - the graph
**              size - size of graph
**		H - the second graph
**		h_size - size of H
**		edges - storage to return the edges e,f and g 
** Return Value: 1 if G ~ H+e-f-g otherwise 0
** Side Effects: updates edge
** Created:     Kerri Morgan    6th June, 2007
** Modified:	Zoe Bukovac		17th April 2011
				Removed fixed size of graphs
** Comments:
** Acknowledgements:
******************************************************************************/
int isIso_H_efg_plus(int **G, int size, int **H, int h_size, int edges[3][2]) 
{
	int i, j, k, l, m, n;

	for(i=0; i<3; i++)
	{
		for(j=0; j<2; j++)
			edges[i][j]=-1;
	}
	if(size != h_size)
		return 0;

	/* for each edge in graph */
	for(i=0; i<size; i++)
	{
		for(j=i+1; j<size; j++)
		{
			/* e in E(H) */
			if(H[i][j]==1)
			{
				for(k=0; k<size; k++)
				{
					for(l=k+1; l<size; l++)
					{
						if(H[k][l]==0) /* f */
						{
							n=l+1;
							while(n<size)
							{
								if(H[k][n]==0) /* g */
								{
									H[i][j]=0;
									H[j][i]=0;
									H[k][l]=1;
									H[l][k]=1;
									H[k][n]=1;
									H[n][k]=1;
									if (isIso(G, H, size))
									{
									printf("T");
										H[i][j]=1;
										H[j][i]=1;
										H[k][l]=0;
										H[l][k]=0;
										H[k][n]=0;
										H[n][k]=0;
										edges[0][1]=i;
										edges[0][0]=j;
										edges[1][1]=k;
										edges[1][0]=l;
										edges[2][1]=k;
										edges[2][0]=n;

										return 1;
									}
									else
									{
										H[i][j]=1;
										H[j][i]=1;
										H[k][l]=0;
										H[l][k]=0;
										H[k][n]=0;
										H[n][k]=0;
									}

								}
								n++;
							}
							for(m=k+1; m<size; m++)
							{
								for(n=m+1; n<size; n++)
								{
									if(H[m][n]==0) /* g */
									{
										H[i][j]=0;
										H[j][i]=0;
										H[k][l]=1;
										H[l][k]=1;
										H[m][n]=1;
										H[n][m]=1;
										if (isIso(G, H, size))
										{
									printf("T");
											H[i][j]=1;
											H[j][i]=1;
											H[k][l]=0;
											H[l][k]=0;
											H[m][n]=0;
											H[n][m]=0;
											edges[0][1]=i;
											edges[0][0]=j;
											edges[1][1]=k;
											edges[1][0]=l;
											edges[2][1]=m;
											edges[2][0]=n;
											return 1;
										}
										else
										{
											H[i][j]=1;
											H[j][i]=1;
											H[k][l]=0;
											H[l][k]=0;
											H[m][n]=0;
											H[n][m]=0;
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}
	return 0;
}

/******************************************************************************
** Function:   isIso_H_k_minus_l_plus
** Description: Function checks if G is isomorphic to H plus 
**		l additional edges and minus k other edges
** Inputs:      G - the graph
**		Gx - the altered graph (initially G)
**              size - size of graph
**		H - the second graph
**		h_size - size of H
**		add, rem - number to add/remove
** Return Value: 1 if G ~ Gx
** Side Effects: Gx is G with some edges removed and others added
** Created:     Kerri Morgan    13th August, 2007
** Modified:	Zoe Bukovac		17th April 2011
				Removed fixed size of graphs
** Comments:	Initial Call to this function has G=Gx
** Acknowledgements:
******************************************************************************/
int isIso_H_k_minus_l_plus(int **G, int **Gx, int size, int **H, int h_size, int add, int rem)
{
	int i, j; 
	if(add<0)
	{
		return 0;
	}
	else if(rem<0)
	{
		return 0;
	}
	if (isIso(Gx, H, size))
	{
		return 1;
	}
	if(add+rem==0)
	{
		return 0;
	}
	for(i=0; i<size; i++)
	{
		for(j=i+1; j<size; j++)
		{
			if(G[i][j]==0) /* not in G */
			{
				if((add>0) &&(Gx[i][j]==0)) /* not in Gx d can add another edge*/
				{
					add--;
					Gx[i][j]=1;
					Gx[j][i]=1;
					if(isIso(Gx, H, size))
					{
						return 1;
					}
					if(isIso_H_k_minus_l_plus(G, Gx, size, H, h_size, add, rem)==1)
					{
						return 1;
					}
					add++;
					Gx[i][j]=0;
					Gx[j][i]=0;
				}
			}
			if(G[i][j]==1) /* edge in G */
			{
				if((rem>0) && (Gx[i][j]==1)) /* in Gx and can remove another edge*/
				{
					rem--;
					Gx[i][j]=0;
					Gx[j][i]=0;
					if(isIso(Gx, H, size))
					{
						return 1;
					}
					if(isIso_H_k_minus_l_plus(G, Gx, size, H, h_size, add, rem)==1)
					{
						return 1;
					}
					rem++;
					Gx[i][j]=1;
					Gx[j][i]=1;
				}
			}
		}
	}
	return 0;
}
/******************************************************************************
** Function:   isIso_H_efg_minus
** Description: Function checks if G is isomorphic to H+e-f-g where
**		e is not in H2 and f and g are
** Inputs:      G - the graph
**              size - size of graph
**		H - the second graph
**		h_size - size of H
**		edges - storage to return the edges e,f and g 
** Return Value: 1 if G ~ H+e-f-g otherwise 0
** Side Effects: updates edge
** Created:     Kerri Morgan    6th June, 2007
** Modified:	Zoe Bukovac		17th April 2011
				Removed fixed size of graphs
** Comments:
** Acknowledgements:
******************************************************************************/
int isIso_H_efg_minus(int **G, int size, int **H, int h_size, int edges[3][2]) 
{
	int i, j, k, l, m, n;

	for(i=0; i<3; i++)
	{
		for(j=0; j<2; j++)
			edges[i][j]=-1;
	}
	if(size != h_size)
		return 0;

	/* for each edge in graph */
	for(i=0; i<size; i++)
	{
		for(j=i+1; j<size; j++)
		{
			/* e not in E(H) */
			if(H[i][j]!=1)
			{
				for(k=0; k<size; k++)
				{
					for(l=k+1; l<size; l++)
					{
						if(H[k][l]==1) /* f */
						{
							n=l+1;
							while(n<size)
							{
								if(H[k][n]==1) /* g */
								{
									H[i][j]=1;
									H[j][i]=1;
									H[k][l]=0;
									H[l][k]=0;
									H[k][n]=0;
									H[n][k]=0;
									if (isIso(G, H, size))
									{
										H[i][j]=0;
										H[j][i]=0;
										H[k][l]=1;
										H[l][k]=1;
										H[k][n]=1;
										H[n][k]=1;
										edges[0][1]=i;
										edges[0][0]=j;
										edges[1][1]=k;
										edges[1][0]=l;
										edges[2][1]=k;
										edges[2][0]=n;

										return 1;
									}
									else
									{
										H[i][j]=0;
										H[j][i]=0;
										H[k][l]=1;
										H[l][k]=1;
										H[k][n]=1;
										H[n][k]=1;
									}

								}
								n++;
							}
							for(m=k+1; m<size; m++)
							{
								for(n=m+1; n<size; n++)
								{
								//displayGraph(H, size);
									if(H[m][n]==1) /* g */
									{
										H[i][j]=1;
										H[j][i]=1;
										H[k][l]=0;
										H[l][k]=0;
										H[m][n]=0;
										H[n][m]=0;
										if (isIso(G, H, size))
										{
									//displayGraph(H, size);
											H[i][j]=0;
											H[j][i]=0;
											H[k][l]=1;
											H[l][k]=1;
											H[m][n]=1;
											H[n][m]=1;
											//printf("%d %d %d %d %d %d\n", i, j, k, l, m, n);
											edges[0][1]=i;
											edges[0][0]=j;
											edges[1][1]=k;
											edges[1][0]=l;
											edges[2][1]=m;
											edges[2][0]=n;
											return 1;
										}
										else
										{
											H[i][j]=0;
											H[j][i]=0;
											H[k][l]=1;
											H[l][k]=1;
											H[m][n]=1;
											H[n][m]=1;
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}
	return 0;
}
	
/******************************************************************************
** Function:   isIso_H_minus_e
** Description: Function checks if G is isomorphic to H-e where e is some edge
** Inputs:      G - the graph
**              size - size of graph
**		H - the second graph
**		h_size - size of H
**		edge - storage to return the edge e such that G~H-e
** Return Value: 1 if G ~ H-e otherwise 0
** Side Effects: updates edge
** Created:     Kerri Morgan    9th May, 2007
** Modified:	Zoe Bukovac		17th April 2011
				Removed fixed size of graphs
** Comments:
** Acknowledgements:
******************************************************************************/
int isIso_H_minus_e(int **G, int size, int **H, int h_size, int edge[2]) 
{
	int i, j;

	edge[0]=-1;
	edge[1]=-1;

	if(size != h_size)
		return 0;

	/* for each edge in graph */
	for(i=0; i<size; i++)
	{
		for(j=i+1; j<size; j++)
		{
			if(H[i][j]==1)
			{
				/* remove it from the graph */
				H[i][j]=0;
				H[j][i]=0;

				if (isIso(G, H, size))
				{
					H[i][j]=1;
					H[j][i]=1;
					edge[0]=i;
					edge[1]=j;
					return 1;
				}
				else
				{
					H[i][j]=1;
					H[j][i]=1;
				}
			}
		}
	}
	return 0;
}
